import pandas as pd
import socket

from datetime import datetime, time
from dotenv import load_dotenv
from io import StringIO
from os import getenv

load_dotenv()

daily_columns = ['Date', 'Open', 'High', 'Low', 'Close', 'Volume']
intraday_columns = ['Date', 'Time', 'Open', 'High', 'Low', 'Close', 'Volume']
iqfeed_columns = ['Date', 'High', 'Low', 'Open', 'Close', 'Total Volume', 'Period Volume']

def read_data_from_socket(sock, recv_buffer = 4096):
    buffer = b''
    data = b''
    while True:
        data = sock.recv(recv_buffer)
        buffer += data

        if b'!ENDMSG!' in buffer:
            break

    # Remove the ENDMSG string
    buffer = buffer[:-12]
    return buffer.decode('utf-8')

def _get_daily_data(symbol: str, max_days: int) -> pd.DataFrame:
    message = 'HDX,{symbol},{max_days},1\n'.format(
        symbol = symbol,
        max_days = max_days
    )

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((getenv('IQFEED_HOST'), int(getenv('IQFEED_PORT'))))

    sock.sendall(message.encode('utf-8'))
    data = read_data_from_socket(sock)
    sock.close()

    # Remove extraneous line endings
    data = ''.join(data.split('\r'))
    data = data.replace(',\n','\n')[:-1]

    data = StringIO(data)
    return pd.read_csv(data, index_col = False, names = iqfeed_columns)

def get_daily_data(symbol: str, start_date: datetime, end_date: datetime) -> pd.DataFrame:
    max_days = (datetime.today() - start_date).days + 1
    data = _get_daily_data(symbol, max_days)

    # Filter data between start_date and end_date
    data['Date'] = pd.to_datetime(data['Date'])
    data['Date'] = data['Date'].dt.date
    data = data[(data['Date'] >= start_date.date()) & (data['Date'] <= end_date.date())]
    data = data.reset_index(drop = True)

    data = data.rename(columns = { 'Total Volume': 'Volume' })
    return data[daily_columns]

def seconds(value):
    return value

def minutes(value):
    return value * 60

def hours(value):
    return value * 3600

def get_historical_data(symbol: str, interval: int, date: datetime = None,
        start_date: datetime = None, end_date: datetime = None, 
        start_filter_time: datetime = datetime.strptime('093000', '%H%M%S').time(), 
        end_filter_time: datetime = datetime.strptime('160000', '%H%M%S').time()) -> pd.DataFrame:

    if date:
        start_date = date
        end_date = datetime.combine(date, time.max)

    message = 'HIT,{symbol},{interval},{begin},{end},,{begin_filter},{end_filter},1\n'.format(
        symbol = symbol,
        interval = interval,
        begin = start_date.strftime('%Y%m%d %H%M%S'),
        end = end_date.strftime('%Y%m%d %H%M%S') if end_date else '',
        begin_filter = start_filter_time.strftime('%H%M%S') if start_filter_time else '',
        end_filter = end_filter_time.strftime('%H%M%S') if end_filter_time else ''
    )

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((getenv('IQFEED_HOST'), int(getenv('IQFEED_PORT'))))

    sock.sendall(message.encode('utf-8'))
    data = read_data_from_socket(sock)
    sock.close()

    # Remove extraneous line endings
    data = ''.join(data.split('\r'))
    data = data.replace(',\n','\n')[:-1]

    data = StringIO(data)
    data = pd.read_csv(data, index_col = False, names = iqfeed_columns)

    if data[iqfeed_columns].isna().any(axis = 1).any() or data.empty:
        return pd.DataFrame(columns = intraday_columns)

    timestamp = pd.to_datetime(data['Date'])
    data['Date'] = timestamp.dt.date
    data['Time'] = timestamp.dt.time
    data = data.rename(columns = { 'Period Volume': 'Volume' })
    return data[intraday_columns]

if __name__ == '__main__':
    symbols = ['SPY']
    date = datetime(2024, 3, 28)
    for symbol in symbols:
        data = get_historical_data(symbol, minutes(5), date)
        print(data)
